from typing import Optional, List, Type

from bson import ObjectId
from pydantic import Field
from pydantic.main import BaseModel

from beanie.collections import Collection
from beanie.cursor import Cursor
from beanie.exceptions import DocumentAlreadyCreated, DocumentWasNotSaved
from beanie.fields import PydanticObjectId


class Document(BaseModel):
    id: Optional[PydanticObjectId] = Field(None, alias="_id")
    _collection_storage = {}

    @classmethod
    def collection(cls) -> Collection:
        """
        Get the collection, which is associated with this document model

        :return: collection
        """
        return cls._collection_storage.get(cls.__name__, None)

    @classmethod
    def set_collection(cls, collection: Collection):
        cls._collection_storage[cls.__name__] = collection

    def __init__(self, *args, **kwargs):
        """
        Initialization

        :param args:
        :param kwargs:
        """
        super(Document, self).__init__(*args, **kwargs)
        if self.collection() is None:
            raise AttributeError(
                "No collection is associated with this document model"
            )

    async def _sync(self) -> None:
        """
        Update local document from the database
        :return: None
        """
        new_instance = await self.get(self.id)
        for key, value in dict(new_instance).items():
            setattr(self, key, value)

    async def create(self) -> "Document":
        """
        Create the document in the database
        :return: Document
        """
        if self.id is not None:
            raise DocumentAlreadyCreated
        await self.collection().insert_one(self)
        return self

    async def replace(self) -> None:
        """
        Fully update the document in the database
        :return: None
        """
        if self.id is None:
            raise DocumentWasNotSaved
        await self.collection().replace_one(self)

    async def update(
        self, update_query: dict, filter_query: dict = None
    ) -> None:
        """
        Partially update the document in the database

        :param update_query: The modifications to apply.
        :param filter_query: The selection criteria for the update. Optional.
        :return: None
        """
        await self.collection().update_one(
            self, update_query=update_query, filter_query=filter_query
        )
        await self._sync()

    async def delete(self) -> None:
        """
        Delete the document

        :return: None
        """
        await self.collection().delete_one(self)
        self.id = None

    @classmethod
    async def get(cls, document_id) -> "Document":
        """
        Get the document bu id

        :param document_id: Id of the documnt
        :return: Document
        """
        return await cls.collection().get_one(document_id)

    @classmethod
    async def find_one(cls, query: dict) -> "Document":
        """
        Find one document

        :param query: The search criteria
        :return: Document
        """
        return await cls.collection().find_one(query)

    @classmethod
    def all(cls) -> Cursor:
        """
        Get all the documents
        :return: AsyncGenerator of the documents
        """
        return cls.collection().all()

    @classmethod
    def find(cls, query) -> Cursor:
        """
        Find many documents by criteria

        :param query: The search criteria
        :return: List of the documents
        """
        return cls.collection().find(query=query)

    @classmethod
    def aggregate(
        cls, query: List[dict], item_model: Type[BaseModel] = None
    ) -> Cursor:
        """
        Aggregate

        :param query: Query with the aggregation commands
        :param item_model: Model of item to return in the list of aggregations
        :return: AsyncGenerator of aggregated items
        """
        return cls.collection().aggregate(query=query, item_model=item_model)

    class Config:
        json_encoders = {
            ObjectId: lambda v: str(v),
        }
        allow_population_by_field_name = True



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['beanie']

package_data = \
{'': ['*']}

install_requires = \
['motor>=2.1.0,<3.0.0', 'pydantic>=1.5.1,<2.0.0']

setup_kwargs = {
    'name': 'beanie',
    'version': '0.1.0',
    'description': 'Pydantic-based Mongo ORM',
    'long_description': '============\nDescription\n============\n\n\nBeanie - is asynchronous ORM for MongoDB, based on `Motor <https://motor.readthedocs.io/en/stable/>`_ and `Pydantic <https://pydantic-docs.helpmanual.io/>`_.\n\nBeanie uses an abstraction over Pydantic Models Motor collections to work with mongo. Document and Collection classes allow to create, replace, update, get, find and aggregate.\n\nOne collection can be associated with only one Document and it helps to keep it structured.\n\nHere you can see, how to use Beanie in simple examples:\n\n============\nInstallation\n============\n\n----\nPIP\n----\n\n.. code-block:: bash\n\n    pip install beanie\n\n\n------\nPoetry\n------\n\n.. code-block:: bash\n\n    poetry add beanie\n\n============\nUsage\n============\n\n-----\nInit\n-----\n\n.. code-block:: python\n\n    from typing import List\n\n    import motor\n    from pydantic import BaseSettings, BaseModel\n\n    from collections import Collection\n    from documents import Document\n\n    # CREATE MOTOR CLIENT AND DB\n\n    client = motor.motor_asyncio.AsyncIOMotorClient(\n        "mongodb://user:pass@host:27017/db",\n        serverSelectionTimeoutMS=100\n    )\n    db = client.beanie_db\n\n\n    # CREATE BEANIE DOCUMENT STRUCTURE\n\n    class SubDocument(BaseModel):\n        test_str: str\n\n\n    class DocumentTestModel(Document):\n        test_int: int\n        test_list: List[SubDocument]\n        test_str: str\n\n\n    # CREATE BEANIE COLLECTION WITH DocumentTestModel STRUCTURE\n\n    test_collection = Collection(\n        name="test_collection", db=db, document_model=DocumentTestModel\n    )\n\n\n---------\nDocuments\n---------\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^\nCreate a document (Insert)\n^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    document = DocumentTestModel(\n        test_int=42,\n        test_list=[SubDocument(test_str="foo"), SubDocument(test_str="bar")],\n        test_str="kipasa",\n    )\n\n    await document.create()\n\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nReplace the document (full update)\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    document.test_str = "REPLACED_VALUE"\n    await document.replace()\n\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nUpdate the document (partial update)\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nin this example, I\'ll add an item to the document\'s "test_list" field\n\n.. code-block:: python\n\n    to_insert = SubDocument(test_str="test")\n    await document.update(update_query={"$push": {"test_list": to_insert.dict()}})\n\n^^^^^^^^^^^^^^^^^^^^^^^\nGet the document\n^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    document = await DocumentTestModel.get(DOCUMENT_ID)\n\n^^^^^^^^^^^^^^^^^^^^^^^\nFind one document\n^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    document = await DocumentTestModel.find_one({"test_str": "kipasa"})\n\n^^^^^^^^^^^^^^^^^^^^^^^\nFind the documents\n^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    async for document in DocumentTestModel.find({"test_str": "uno"}):\n        print(document)\n\nOR\n\n.. code-block:: python\n\n    documents =  await DocumentTestModel.find({"test_str": "uno"}).to_list()\n\n^^^^^^^^^^^^^^^^^^^^^^^\nGet all the documents\n^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    async for document in DocumentTestModel.all()\n        print(document)\n\nOR\n\n.. code-block:: python\n\n    documents = await DocumentTestModel.all().to_list()\n\n^^^^^^^^^^^^^^^^^^^^^^^\nDelete the document\n^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    await document.delete()\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nAggregate from the document model\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    async for item in DocumentTestModel.aggregate(\n        [{"$group": {"_id": "$test_str", "total": {"$sum": "$test_int"}}}]\n    ):\n        print(item)\n\nOR\n\n.. code-block:: python\n\n    class OutputItem(BaseModel):\n        id: str = Field(None, alias="_id")\n        total: int\n\n    async for item in DocumentTestModel.aggregate(\n        [{"$group": {"_id": "$test_str", "total": {"$sum": "$test_int"}}}],\n        item_model=OutputModel\n    ):\n        print(item)\n\nOR\n\n.. code-block:: python\n\n    results = await DocumentTestModel.aggregate(\n        [{"$group": {"_id": "$test_str", "total": {"$sum": "$test_int"}}}],\n        item_model=OutputModel\n    ).to_list()\n\n\n\n------------\nCollections\n------------\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nInsert the document into the collection\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    inserted_document = await collection.insert_one(document)\n\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nReplace the document\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n     await collection.replace_one(document)\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nUpdate the document\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    to_insert = SubDocument(test_str="test")\n    await collection.update_one(\n        document, update_query={"$push": {"test_list": to_insert.dict()}}\n    )\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nUpdate many documents\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    await collection.update_many(\n        update_query={"$set": {"test_int": 100}}, filter_query={"test_str": "kipasa"},\n    )\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nDelete the document\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    await collection.delete_one(document)\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nGet the document\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    document = await collection.get_one(DOCUMENT_ID)\n\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nFind the document\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    document = await collection.find_one({"test_str": "one"})\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nFind many documents\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    async for document in collection.find({"test_str": "uno"}):\n        print(document)\n\nOR\n\n.. code-block:: python\n\n    documents = await collection.find({"test_str": "uno"}).to_list()\n\nOR\n\n.. code-block:: python\n\n    documents = await collection.find({"test_str": "uno"}).to_list(length=10)\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nGet all the documents from the collection\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    async for document in collection.all():\n        print(document)\n\nOR\n\n.. code-block:: python\n\n    documents = await collection.all().to_list()\n\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nAggregate\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block:: python\n\n    async for item in collection.aggregate(\n        [{"$group": {"_id": "$test_str", "total": {"$sum": "$test_int"}}}]\n    ):\n        print(item)\n\nOR\n\n.. code-block:: python\n\n    class OutputItem(BaseModel):\n        id: str = Field(None, alias="_id")\n        total: int\n\n    async for item in collection.aggregate(\n        [{"$group": {"_id": "$test_str", "total": {"$sum": "$test_int"}}}],\n        item_model=OutputModel\n    ):\n        print(item)\n\nOR\n\n.. code-block:: python\n\n\n    results = await collection.aggregate(\n        [{"$group": {"_id": "$test_str", "total": {"$sum": "$test_int"}}}],\n        item_model=OutputModel\n    ).to_list():\n',
    'author': 'Roman',
    'author_email': 'roman-right@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/roman-right/beanie',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import typing
from typing import Type, List, Union

from motor.motor_asyncio import AsyncIOMotorDatabase
from pydantic import BaseModel
from pymongo.results import UpdateResult, DeleteResult

from beanie.cursor import Cursor
from beanie.exceptions import DocumentNotFound
from beanie.fields import PydanticObjectId

if typing.TYPE_CHECKING:
    from beanie.documents import Document


class Collection:
    def __init__(
        self,
        name: str,
        db: AsyncIOMotorDatabase,
        document_model: Type["Document"],
    ) -> None:
        """
        Collection initialization

        :param name: Collection's name
        :param db: AsyncIOMotorDatabase instance
        :param document_model: Subclass of Document to follow the structure
        """
        self.db = db
        self.document_model = document_model
        self.motor_collection = db[name]

        self.document_model.set_collection(self)

    async def insert_one(self, document: "Document") -> "Document":
        """
        Insert one document into the collection

        :param document: Document
        :return: Inserted document with updated id field
        """
        if not isinstance(document, self.document_model):
            raise TypeError(
                "Document must be object of the Document class"
            )
        result = await self.motor_collection.insert_one(
            document.dict(by_alias=True, exclude={"id"})
        )
        document.id = PydanticObjectId(result.inserted_id)
        return document

    async def replace_one(self, document: "Document") -> UpdateResult:
        """
        Replace already created (stored to the collection) document

        :param document: document
        :return: UpdateResult instance
        """
        if not isinstance(document, self.document_model):
            raise TypeError(
                "Document must be object of the Document class"
            )
        if document.id is None:
            raise ValueError("Document must have id")
        result = await self.motor_collection.replace_one(
            {"_id": document.id}, document.dict(by_alias=True)
        )
        if not result.raw_result["updatedExisting"]:
            raise DocumentNotFound
        return result

    async def update_one(
        self,
        document: "Document",
        update_query: dict,
        filter_query: dict = None,
    ) -> UpdateResult:
        """
        Partially update already created document

        :param document: Document
        :param update_query: The modifications to apply.
        :param filter_query: The selection criteria for the update. Optional.
        :return: UpdateResult instance
        """
        if not isinstance(document, self.document_model):
            raise TypeError(
                "Document must be object of the Document class"
            )
        if filter_query is None:
            return await self.motor_collection.update_one(
                {"_id": document.id}, update_query
            )
        else:
            return await self.motor_collection.update_one(
                filter_query, update_query
            )

    async def update_many(
        self, filter_query: dict, update_query: dict
    ) -> UpdateResult:
        """
        Partially update many documents

        :param filter_query: The modifications to apply.
        :param update_query: The selection criteria for the update.
        :return: UpdateResult instance
        """
        return await self.motor_collection.update_many(
            filter_query, update_query
        )

    async def delete_one(self, document: "Document") -> DeleteResult:
        """
        Delete the document

        :param document: Document to delete
        :return: DeleteResult instance
        """
        if not isinstance(document, self.document_model):
            raise TypeError(
                "Document must be object of the Document class"
            )
        return await self.motor_collection.delete_one({"_id": document.id})

    async def find_one(self, query: dict) -> Union["Document", None]:
        """
        Find one document by criteria

        :param query: The selection criteria
        :return: Document
        """
        document = await self.motor_collection.find_one(query)
        if document is None:
            return None
        return self.document_model(**document)

    async def get_one(
        self, document_id: PydanticObjectId
    ) -> Union["Document", None]:
        """
        Get one document by id

        :param document_id: Id of the document
        :return: Document
        """
        if not isinstance(document_id, PydanticObjectId):
            raise TypeError("Id must be of type PydanticObjectId")
        return await self.find_one({"_id": document_id})

    def find(self, query: dict) -> Cursor:
        """
        Find many documents by criteria

        :param query: The selection criteria.
        :return: AsyncGenerator of the documents
        """
        cursor = self.motor_collection.find(query)
        return Cursor(motor_cursor=cursor, model=self.document_model)

    def all(self) -> Cursor:
        """
        Get all the documents

        :return: AsyncGenerator of the documents
        """
        return self.find(query={})

    def aggregate(
        self, query: List[dict], item_model: Type[BaseModel] = None
    ) -> Cursor:
        """
        Aggregate

        :param query: Query with aggregation commands
        :param item_model: Model of item to return in the list of aggregations
        :return: AsyncGenerator of aggregated items
        """
        cursor = self.motor_collection.aggregate(query)
        return Cursor(motor_cursor=cursor, model=item_model)
